/*
 * Program flash SIMM on Krups and Espresso
 *
 * I may be a good idea to mmap the flash into our space and
 * program it from the userland because no driver would be neeed.
 * But so far we let kernel to work out details.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>

#include <asm/jsflash.h>

typedef struct {
	unsigned long	a_magic;        /* magic number */
	unsigned long	a_text;         /* size of text segment */
	unsigned long	a_data;         /* size of initialized data */
	unsigned long	a_bss;          /* size of uninitialized data */
	unsigned long	a_syms;         /* size of symbol table */
	unsigned long	a_entry;        /* entry point */
	unsigned long	a_trsize;	/* size of text relocation */
	unsigned long	a_drsize;	/* size of data relocation */
} Exec;

#define FLASHPHYS  0x20000000
#define MEG        0x00100000

struct par {
	char *fdev;
	char *img;
	char *kimg;
	int csum;
};

void make_check_sum(char *data, unsigned int dsize);
void parseargs(struct par *, int argc, char **argv);
void Usage(void);
void Error0(int);
void Error1(int, char *);

void
main(int argc, char **argv)
{
	struct par parb;
	int fd, ifd, kfd = -1;
	char *mp, *kp = 0;
	struct stat statb;
	struct jsflash_ident_arg id;
	struct jsflash_program_arg prga;
	unsigned int ileng, total;
	unsigned int kleng = 0;
	Exec ahdr;
	char *es;
	int rc;

	parseargs(&parb, argc, argv);

	if ((fd = open(parb.fdev, O_RDWR)) == -1) {
		Error1(errno, parb.fdev);
	}

	if ((ifd = open(parb.img, O_RDONLY)) == -1) {
		Error1(errno, parb.img);
	}

	/*
	 * Now let's get length.
	 *
	 * PROM gets the top of all segments as size, this may differ
	 * from file length if the image is padded. But normally
	 * they are the same.
	 *

: elf-size
   0 1 do-map-in                ( ... -- elfsig T vaddr )       \ map ELF header
   get-elf32hdr
   e32_shnum e32_shentsize *    ( ... -- elfsig T vaddr shoff )
   e32_shoff
   rot 1 do-map-out             ( elfsig T bytes shoff )
   2dup + >r over
   do-map-in                                    \ map section header
   2dup r> -rot swap bounds ?do
      i >sh32_type l@ 8 <> if                   \ Type not NOBITS
         i >sh32_size l@ i >sh32_offset l@ +    \ ( ... -- ... top_of_seg )
         i >sh32_addralign l@
         round-up
         umax
      then
   e32_shentsize +loop
   >r swap do-map-out r> /l + 16meg umin /l umax
;

	 */
	if (fstat(ifd, &statb) == -1) {
		Error1(errno, parb.img);
	}
	if (statb.st_size > 4*MEG - 8) {
		/*
		 * Minimum size is 4MB, Sun does not deliver that size.
		 * Common size is 8MB, and maximum is 16MB.
		 * Due to some weird design, involving a memory hole,
		 * PROM offsets 8MB SIMM by 4M.
		 * We do not have a self-tuning code so far,
		 * and we limit the size.
		 */
		fprintf(stderr,
		    "jsflash: cannot program more than %d(0x%x) bytes\n",
		    4*MEG, 4*MEG);
		exit(1);
	}
	ileng = statb.st_size;

	/*
	 * We must program 4 bytes at a time due to schematics of the SIMM.
	 * Also let's make room for the checksum.
	 */
	total = ((ileng + sizeof(int)) + 3) & (~3);
	if ((mp = malloc(total)) == NULL) {
		fprintf(stderr, "jsflash: no core\n");
		exit(1);
	}

	if ((rc = read(ifd, mp, ileng)) < 0) {
		Error1(errno, parb.img);
	}
	/* Since we get length from stat(), no need to check rc<ileng. */
	memset(mp+ileng, 0, total-ileng);
	if (parb.csum) {
		make_check_sum(mp, ileng);
	}

	/*
	 * We open and preload kernel before we program the bootstrap
	 * in order not to erase flash in case of a typo in a command line.
	 */
	if (parb.kimg != 0) {
		if ((kfd = open(parb.kimg, O_RDONLY)) == -1) {
			Error1(errno, parb.kimg);
		}

		if (read(kfd, (char*)&ahdr, sizeof(Exec)) < sizeof(Exec)) {
			fprintf(stderr,
			    "jsfprog: Cannot read header from %s\n", parb.kimg);
			exit(1);
		}

		kleng = sizeof(Exec);
		kleng += ((ahdr.a_text + 3) & (~3));
		kleng += ((ahdr.a_data + 3) & (~3));

		if ((kp = malloc(kleng)) == NULL) {
			fprintf(stderr, "jsflash: no core\n");
			exit(1);
		}

		lseek(kfd, 0, SEEK_SET);
		if (read(kfd, kp, kleng) < kleng) {
			fprintf(stderr,
			    "jsfprog: Cannot read kernel %s\n", parb.kimg);
			exit(1);
		}
	}

	if (ioctl(fd, JSFLASH_IDENT, &id) != 0) {
		if (errno == ENOTTY) {
			fprintf(stderr,
			    "jsflash: %s: Not a JavaStation Flash\n",
			    parb.fdev);
		} else {
			if ((es = strerror(errno)) == NULL) es = "???";
			fprintf(stderr, "jsflash: Cannot probe %s: %s\n",	
			    parb.fdev, es);
		}
		exit(1);
	}
	printf("Found \"%s\" at 0x%lx[%x]\n", id.name, (long)id.off, id.size);

	printf("Erasing...\n");
	fflush(stdout);		/* Keep kernel messages from overlappting */
	if (ioctl(fd, JSFLASH_ERASE, 0) != 0) {
		Error1(errno, "JSFLASH_ERASE");
	}

	printf("Programming %d(0x%x) bytes...\n", (int)ileng, (int)ileng);
	prga.off = 0x20400000;		/* XXX Use id.off here somehow */
	prga.size = total;
	prga.data = (unsigned long) mp;
	if (ioctl(fd, JSFLASH_PROGRAM, &prga) != 0) {
		Error1(errno, "JSFLASH_PROGRAM");
	}

	if (parb.kimg != 0) {
		printf("Kernel %d(0x%x) bytes...\n", (int)kleng, (int)kleng);
		prga.off = 0x20500000;		/* XXX */
		prga.size = kleng;
		prga.data = (unsigned long) kp;
		if (ioctl(fd, JSFLASH_PROGRAM, &prga) != 0) {
			Error1(errno, "JSFLASH_PROGRAM");
		}
	}

	if (parb.kimg != 0) {
		free(kp);
		close(kfd);
	}
	free(mp);
	close(fd);
	close(ifd);
	exit(0);
}

/*
 * Append checksum to the data at offset dsize.
 *
: checksum-ok?		( flashp elfsize -- ? )
   /l - 2>r
   0			( sum=0 )
   2r@ bounds ?do
      i c@ +		\ UNSIGNED on SPARC
   loop
   2r>
   + unaligned-l@ =
;
 */
void make_check_sum(char *data, unsigned int dsize)
{
	unsigned int sum;

	sum = 0;
	while (dsize-- != 0) {
		sum += *((unsigned char *)data)++;
	}
	data[0] = sum >> 24;
	data[1] = sum >> 16;
	data[2] = sum >> 8;
	data[3] = sum;
}

void parseargs(struct par *p, int argc, char **argv)
{
	char *arg;

	--argc;  argv++;

	p->fdev = "/dev/jsflash";
	p->csum = 0;
	p->img = 0;	/* in case */
	p->kimg = 0;
	while ((arg = *argv) != 0 && arg[0] == '-') {
		--argc;  argv++;
		if (arg[1] == 'c') {
			p->csum = 1;
		} else if (arg[1] == 'd') {
			if ((arg = *argv) == 0) Usage();
			p->fdev = arg;
			--argc;  argv++;
		} else if (arg[1] == 'k') {
			if ((arg = *argv) == 0) Usage();
			p->kimg = arg;
			--argc;  argv++;
		} else {
			Usage();
		}
	}
	if (arg == 0 || argv[1] != 0) Usage();
	p->img = arg;
}

void Usage() {
	fprintf(stderr,
	    "Usage: jsflash [-c] [-d fdev] [-k kernel] bootstrap\n");
	exit(1);
}

/*
 * I am too lazy to work with varargs/stdarg properly.
 */
void Error1(int err, char *s0) {
	char *es;
	if ((es = strerror(err)) == NULL) es = "???";
	fprintf(stderr, "jsflash: %s: %s\n", s0, es);
	exit(1);
}
